html, body {
    background-color: #303538;
    color: white;
    padding: 0;
    margin: 0;
}

.margins {
    padding: 0 30px;
}
label {
    user-select: none;
    * {
        user-select: initial;
    }
}
